/*
 * Copyright (c) 2020 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef _ICUB_H
#define _ICUB_H

#include <stdbool.h>
#include <stdint.h>

/* Opaque structures */
struct icub_dev;
struct icub_node;

/* forward declaration */
struct strain2_data;


/* Known board types - page 7 */
/* only listing types managed by this library here */
#define ICUB_BOARD_STRAIN 0x6
#define ICUB_BOARD_STRAIN2 0xc
#define ICUB_BOARD_UNKNOWN 0xff	/* special value to mark unknown type */

/* icub.c */
extern struct icub_dev* icub_init(const char *);
extern void icub_end(struct icub_dev *);
extern int icub_probe(struct icub_dev *);
extern struct icub_dev* icub_dev(const struct icub_node *);
extern struct icub_node *icub_find_node(struct icub_dev *, int);
extern struct icub_node *icub_get_node(struct icub_dev *, int);

extern int icub_can_address(const struct icub_node *);
extern void icub_dump(const struct icub_dev *);

/* decode.c */
extern int icub_receive_and_decode(struct icub_dev *, int);

/* cmds.c */
extern void icub_sync(struct icub_node *);
extern int icub_discover(struct icub_dev *);
extern int icub_get_additional_info(struct icub_node *);
extern int icub_set_additional_info(struct icub_node *, char [32]);
extern int icub_sensor_set_board_adx(struct icub_node *, uint8_t);
extern int icub_sensor_get_firmware_version(struct icub_node *);
extern int icub_sensor_set_can_data_rate(struct icub_node *, uint8_t);
extern int icub_sensor_set_resolution(struct icub_node *, uint8_t);
extern int icub_sensor_get_fullscale(struct icub_node *, int, int);
extern int icub_sensor_get_adc(struct icub_node *, uint8_t, uint8_t);

extern int icub_sensor_imu_transmit(struct icub_node *, int);


/* TX modes - page 38 */
enum icub_strain_mode {
	ICUB_STRAIN_MODE_TX_CALIBRATED = 0,
	ICUB_STRAIN_MODE_TX_NONE = 1,
	ICUB_STRAIN_MODE_TX_UNCALIBRATED = 3,
	ICUB_STRAIN_MODE_TX_ALL = 4
};
extern int icub_sensor_set_txmode(struct icub_node *,
    enum icub_strain_mode);

/* IMU Conf register - page 39 */
enum icub_imu_sensor_mask  {
	ICUB_IMU_SENSOR_MASK_NONE = 0x0,
	ICUB_IMU_SENSOR_MASK_ACC = 0x0001,
	ICUB_IMU_SENSOR_MASK_MAG = 0x0002,
	ICUB_IMU_SENSOR_MASK_GYR = 0x0004,
	ICUB_IMU_SENSOR_MASK_EUL = 0x0008,
	ICUB_IMU_SENSOR_MASK_QUA = 0x0010,
	ICUB_IMU_SENSOR_MASK_LIA = 0x0020,
	ICUB_IMU_SENSOR_MASK_GRV = 0x0040,
	ICUB_IMU_SENSOR_MASK_STATUS = 0x8000
};
extern int icub_sensor_imu_config_set(struct icub_node *, uint16_t);
extern int icub_sensor_imu_config_get(struct icub_node *);

/* IMU calibration status - page 93 */
enum icub_imu_calib_status {
	ICUB_CALIB_NONE = 0,
	ICUB_CALIB_POOR = 1,
	ICUB_CALIB_MEDIUM = 2,
	ICUB_CALIB_GOOD = 3
};

/* Thermometer masks - page 42 */
enum icub_thermometer_mask {
	ICUB_THERMOMETER_NONE = 0x0,
	ICUB_THERMOMETER_ONE = 0x1
};

extern int icub_sensor_thermometer_config_set(struct icub_node *,
    enum icub_thermometer_mask);
extern int icub_sensor_thermometer_config_get(struct icub_node *);
extern int icub_sensor_thermometer_transmit(struct icub_node *, int);

/* Regulation sets - page 49 */
enum icub_regset {
	ICUB_REGSET_INUSE = 0,
	ICUB_REGSET_FIRST = 1,
	ICUB_REGSET_SECOND = 2,
	ICUB_REGSET_THIRD = 3
};

enum icub_regset_validity {
	ICUB_REGSET_VALIDITY_TEMPORARY = 0,
	ICUB_REGSET_VALIDITY_PERMANENT = 1
};
extern int icub_regulationset_set(struct icub_node *, enum icub_regset,
    enum icub_regset_validity);
extern int icub_regulationset_get(struct icub_node *,
    enum icub_regset_validity);

extern int icub_set_matrix_rc(struct icub_node *,
    enum icub_regset, uint8_t, uint8_t, float);
extern int icub_get_matrix_rc(struct icub_node *,
    enum icub_regset, uint8_t, uint8_t);
extern int icub_set_fullscale(struct icub_node *,
    enum icub_regset, uint8_t, uint16_t);
extern int icub_set_serial_number(struct icub_node *, char *);
extern int icub_get_serial_number(struct icub_node *);
extern int icub_set_ch_dac(struct icub_node *,
    enum icub_regset, uint8_t, uint16_t);
extern int icub_get_ch_dac(struct icub_node *, enum icub_regset, uint8_t);

enum icub_calib_mode {		/* page 58 */
	ICUB_CALIB_MODE_EVERY_CHANNEL_RESET = 0,
	ICUB_CALIB_MODE_EVERY_CHANNEL_NEGATIVE_OF_DAC = 1,
	ICUB_CALIB_MODE_SET_CHANNEL_WITH_VALUE = 2
};
extern int icub_set_calib_tare(struct icub_node *,
    enum icub_calib_mode, enum icub_regset, uint8_t, float);
extern int icub_get_calib_tare(struct icub_node *,
    enum icub_regset, uint8_t);
extern int icub_get_eeprom_status(struct icub_node *);
extern int icub_save2ee(struct icub_node *);
extern int icub_amplifier_reset(struct icub_node *, enum icub_regset, uint8_t);
extern int icub_amplifier_range_of_gain_get(struct icub_node *,
    enum icub_regset, uint8_t);
extern int icub_amplifier_range_of_offset_get(struct icub_node *,
    enum icub_regset, uint8_t);
extern int icub_amplifier_gainoffset_get(struct icub_node *,
    enum icub_regset, uint8_t);
extern int icub_amplifier_gainoffset_set(struct icub_node *,
    enum icub_regset, uint8_t, uint8_t, float, uint16_t);
extern int icub_amplifier_offset_autocalib(struct icub_node *,
    enum icub_regset, uint8_t, uint16_t, uint16_t, uint8_t);

/* data.c */
extern int icub_get_strain2_data(struct icub_node *, struct strain2_data *);
extern void icub_print_strain2_config(struct icub_node *);
extern void icub_print_strain2_data(struct icub_node *);

/*
 * Data structure for strain2 nodes
 */
struct strain2_data {
	char serial[8];
	bool eeprom_equal;
	bool eeprom_ok;
	/* Force/torque */
	uint8_t mode;		/* RAW 0 or CALIBRATED 1 */
	float force[3];
	float torque[3];
	float gain[6];
	int16_t offset[6];
	float tare[6];
	float matrix[6][6];
	float fullscale[6];
	uint8_t force_saturation;
	uint8_t torque_saturation;
	float gain_min[6];
	float gain_max[6];
	uint16_t offset_min[6];
	uint16_t offset_max[6];
	uint8_t chmsk;
	uint8_t algmsk;
	uint8_t memsk;
	uint16_t mae;
	enum icub_regset regset;
	enum icub_regset_validity validity;
	/* IMU */
	uint8_t seq;		/* current sequence number */
	uint16_t imu_config_mask;
	uint8_t imu_fusion_mode;
	float accel[3];		/* m/s^2 */
	float mag[3];		/* micro Tesla */
	float gyro[3];		/* deg/s */
	float euler[3];		/* deg */
	float lia[3];		/* m/s^2 */
	float grv[3];		/* m/s^2 */
	float q[4];		/* quaternion */
	enum icub_imu_calib_status gyr_calib;
	enum icub_imu_calib_status acc_calib;
	enum icub_imu_calib_status mag_calib;
	uint32_t acq_duration;
	/* Thermometer */
	float temperature;	/* celsius degrees */
	int period;             /* ms */
};


#endif	/* _ICUB_H */
