/*
 * Copyright (c) 2020-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <icub/icub.h>
#include "socketcan.h"
#include "icub_protocol.h"
#include "internal.h"
#include "q15.h"

static int
send_command(struct icub_node *node, CAN_MSG *msg, bool reply)
{
	int rv;

	pthread_mutex_lock(&node->mutex);
	while (node->reply_pending)
		pthread_cond_wait(&node->cond, &node->mutex);
	rv = node->parent->send(node->parent->fd, msg);
	if (rv == 0)
		node->reply_pending = reply;
	pthread_mutex_unlock(&node->mutex);
	return rv;
}

/**
 * Wait for a reply
 */
void
icub_sync(struct icub_node *node)
{
	pthread_mutex_lock(&node->mutex);
	while (node->reply_pending)
		pthread_cond_wait(&node->cond,&node->mutex);
	pthread_mutex_unlock(&node->mutex);
}

/**
 * Discover
 *
 * page 19
 */
int
icub_discover(struct icub_dev *dev)
{
	CAN_MSG msg;

	/* send to everyone */
	msg.id = ICUB_MSG_ID(ICUB_CLS_BOOTLOADER, 0, ICUB_NODE_EVERYONE);
	msg.len = 1;
	msg.data[0] = ICUB_CMD_BROADCAST;
	return dev->send(dev->fd, &msg);
}

/**
 * Get Additional info
 *
 * page 20
 */
int
icub_get_additional_info(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_BOOTLOADER, 0, node->addr);
	msg.len = 1;
	msg.data[0] = ICUB_CMD_GET_ADDITIONAL_INFO;
	return send_command(node, &msg, true);
}

/**
 * Set Additional info
 *
 * page 21
 */
int
icub_set_additional_info(struct icub_node *node, char info[32])
{
	CAN_MSG msg;
	int i;

	msg.id = ICUB_MSG_ID(ICUB_CLS_BOOTLOADER, 0, node->addr);
	msg.len = 6;
	msg.data[0] = ICUB_CMD_SET_ADDITIONAL_INFO;
	for (i = 0; i < 8; i++) {
		msg.data[1] = i;
		memcpy(msg.data + 2, info + 4*i, 4);
		if (send_command(node, &msg, false) != 0)
			return -1;
	}
	return 0;
}

/**
 * Set board CAN address
 *
 * page 32
 */
int
icub_sensor_set_board_adx(struct icub_node *node, uint8_t address)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_SET_BOARD_ADX;
	msg.data[1] = address;
	return send_command(node, &msg, false);
}

/**
 * Get firmware version
 *
 * page 33
 */
int
icub_sensor_get_firmware_version(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 3;
	msg.data[0] = ICUB_GET_FIRMWARE_VERSION;
	msg.data[1] = ICUB_CAN_PROTO_MAJOR;
	msg.data[2] = ICUB_CAN_PROTO_MINOR;
	return send_command(node, &msg, true);
}

/**
 * Set CAN data rate
 *
 * page 34
 */
int
icub_sensor_set_can_data_rate(struct icub_node *node, uint8_t rate)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_SET_CANDATARATE;
	msg.data[1] = rate;
	return send_command(node, &msg, false);
}

/**
 * Set resolution
 *
 * page 35
 *
 * Deprecated
 */
int
icub_sensor_set_resolution(struct icub_node *node, uint8_t res)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_SET_RESOLUTION;
	msg.data[1] = res;
	return send_command(node, &msg, false);
}

/**
 * Get full scale
 *
 * page 36
 */
int
icub_sensor_get_fullscale(struct icub_node *node, int rs, int channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_GET_FULLSCALE;
	msg.data[1] = (rs & 0x0f) << 4 | (channel & 0x0f);
	return send_command(node, &msg, true);
}

/**
 * Get ADC
 *
 * page 37
 */
int
icub_sensor_get_adc(struct icub_node *node,
    uint8_t channel, uint8_t mode)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 3;
	msg.data[0] = ICUB_GET_ADC;
	msg.data[1] = channel;
	msg.data[2] = mode;
	return send_command(node, &msg, true);
}

/**
 * Set TX Mode
 *
 * page38
 */
int
icub_sensor_set_txmode(struct icub_node *node,
    enum icub_strain_mode mode)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_SET_TXMODE;
	msg.data[1] = mode;
	node->strain2.mode = mode;
	return send_command(node, &msg, false);
}

/**
 * Set the configuration of the IMU
 *
 * page 39
 */
int
icub_sensor_imu_config_set(struct icub_node *node, uint16_t mask)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 8;
	msg.data[0] = ICUB_IMU_CONFIG_SET;
	msg.data[1] = mask & 0xff;
	msg.data[2] = (mask & 0xff00) >> 8;
	msg.data[3] = 1;	/* FUS */
	msg.data[4] = 0;
	msg.data[5] = 0;
	msg.data[6] = 0;
	msg.data[7] = 0;
	return send_command(node, &msg, false);
}

/**
 * Get the configuration of the IMU
 *
 * page 40
 */
int
icub_sensor_imu_config_get(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 1;
	msg.data[0] = ICUB_IMU_CONFIG_GET;
	return send_command(node, &msg, true);
}

/**
 * Tranmit imu data
 *
 * page 41
 */
int
icub_sensor_imu_transmit(struct icub_node *node, int rate)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_IMU_TRANSMIT;
	msg.data[1] = rate;
	return send_command(node, &msg, false);
}

/**
 * Set Thermometer configuration
 *
 * page 42
 */
int
icub_sensor_thermometer_config_set(struct icub_node *node,
    enum icub_thermometer_mask mask)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_THERMOMETER_CONFIG_SET;
	msg.data[1] = mask;
	return send_command(node, &msg, false);
}

/**
 * Get thermomether config
 *
 * page 43
 */
int
icub_sensor_thermometer_config_get(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 1;
	msg.data[0] = ICUB_THERMOMETER_CONFIG_GET;
	return send_command(node, &msg, true);
}

/**
 * Tranmit thermometer data
 *
 * page 44
 */
int
icub_sensor_thermometer_transmit(struct icub_node *node, int rate)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_THERMOMETER_TRANSMIT;
	msg.data[1] = rate;
	return send_command(node, &msg, false);
}

/**
 * Impose the regulation set of the strain2 board
 *
 * page 49
 */
int
icub_regulationset_set(struct icub_node *node,
    enum icub_regset rs, enum icub_regset_validity validity)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_REGULATIONSET_SET;
	msg.data[1] = rs << 4 | validity;
	return send_command(node, &msg, false);
}

/**
 * Retrieve the regulation set of the strain2 board
 *
 * page 50
 */
int
icub_regulationset_get(struct icub_node *node,
    enum icub_regset_validity validity)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_REGULATIONSET_GET;
	msg.data[1] = validity;
	return send_command(node, &msg, true);
}

/**
 * change te value of the transformation matrix in srain2 board
 *
 * page 51
 */
int
icub_set_matrix_rc(struct icub_node *node,
    enum icub_regset rs, uint8_t row, uint8_t col, float value)
{
	CAN_MSG msg;
	uint16_t q;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 5;
	msg.data[0] = ICUB_SET_MATRIX_RC;
	msg.data[1] = rs << 4 | (row & 0x0f);
	msg.data[2] = col & 0x0f;
	q = float_to_u16(value);
	msg.data[3] = (q & 0xff00) >> 8;
	msg.data[4] = q & 0xff;
	return send_command(node, &msg, false);
}

/**
 * retrieve te value of the transformation matrix in srain2 board
 *
 * page 52
 */
int
icub_get_matrix_rc(struct icub_node *node,
    enum icub_regset rs, uint8_t row, uint8_t col)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 3;
	msg.data[0] = ICUB_GET_MATRIX_RC;
	msg.data[1] = rs << 4 | (row & 0x0f);
	msg.data[2] = col & 0x0f;
	return send_command(node, &msg, true);
}

/**
 * Impose the fullscale of the regulation set of strain2
 *
 * Page 53
 */
int
icub_set_fullscale(struct icub_node *node,
    enum icub_regset rs, uint8_t channel, uint16_t fullscale)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 4;
	msg.data[0] = ICUB_SET_FULLSCALE;
	msg.data[1] = rs << 4 | (channel & 0x0f);
	msg.data[2] = (fullscale & 0xff00) >> 8; /* big endian */
	msg.data[3] = fullscale & 0xff;
	return send_command(node, &msg, false);
}

/**
 * Impose the serial number that specifiles the regulation set
 * of the given strain2 board
 *
 * page 54
 */
int
icub_set_serial_number(struct icub_node *node, char *sn)
{
	CAN_MSG msg;
	int i;

	if (strlen(sn) > 7)
		return -1;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 1 + strlen(sn);
	msg.data[0] = ICUB_SET_SERIAL_NO;
	for (i = 0; i < strlen(sn); i++)
		msg.data[1+i] = sn[i];

	return send_command(node, &msg, false);
}

/**
 * retrieve the serial number that specifies the regulation set of the
 * given strain2 board.
 *
 * page 55
 */
int
icub_get_serial_number(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 1;
	msg.data[0] = ICUB_GET_SERIAL_NO;

	return send_command(node, &msg, true);
}

/**
 * impose an offset in the analog front end of strain/strain2 boards, 
 * so that the acquired ADC values are shifted down or up in value.
 *
 * page 56
 */
int
icub_set_ch_dac(struct icub_node *node,
    enum icub_regset rs, uint8_t channel, uint16_t offset)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 4;
	msg.data[0] = ICUB_SET_CH_DAC;
	msg.data[1] = (rs << 4) | (channel & 0x0f);
	msg.data[2] = (offset & 0xff00) >> 8;
	msg.data[3] = offset & 0xff;

	return send_command(node, &msg, false);
}

/**
 * retrieve the offset in the analog front end of strain2 boards.
 *
 * page 57
 */
int
icub_get_ch_dac(struct icub_node *node,
    enum icub_regset rs, uint8_t channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 3;
	msg.data[0] = ICUB_GET_CH_DAC;
	msg.data[1] = (rs << 4) | (channel & 0x0f);

	return send_command(node, &msg, true);
}

/**
 * impose the tare (digital offset) of the regulation set of strain2 boards.
 *
 * page 58
 */
int
icub_set_calib_tare(struct icub_node *node,
    enum icub_calib_mode mode, enum icub_regset rs, uint8_t channel,
    float tare)
{
	CAN_MSG msg;
	uint16_t q;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.data[0] = ICUB_SET_CALIB_TARE;
	msg.data[1] = mode;
	msg.data[2] = (rs << 4) | (channel & 0x0f);
	if (mode == ICUB_CALIB_MODE_SET_CHANNEL_WITH_VALUE) {
		msg.len = 5;
		q = float_to_q15(tare);
		msg.data[3] = (q & 0xff00) >> 8;
		msg.data[4] = q & 0xff;
	} else
		msg.len = 3;

	return send_command(node, &msg, false);
}	

/**
 * retrieve a value of the calibration tare stored the regulation set
 * of strain2 boards.
 *
 * page 59
 */
int
icub_get_calib_tare(struct icub_node *node,
    enum icub_regset rs, uint8_t channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.data[0] = ICUB_GET_CALIB_TARE;
	msg.data[1] = (rs << 4) | (channel & 0x0f);
	msg.len = 2;

	return send_command(node, &msg, true);
}	

/**
 * Used in strain2 boards to know if the RAM copy of the FT settings
 * is equal to the one in EEPROM.
 *
 * page 60
 */
int
icub_get_eeprom_status(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 1;
	msg.data[0] = ICUB_GET_EEPROM_STATUS;

	return send_command(node, &msg, true);
}

/**
 * Used in strain2 boards to save in permanent storage all RAM data
 * belonging to the FT settings.
 *
 * page 61
 */
int
icub_save2ee(struct icub_node *node)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 1;
	msg.data[0] = ICUB_SAVE2EE;

	return send_command(node, &msg, true);
}

/**
 * agnostic configuration of the gain and offset of the amplifier.
 *
 * page 62
 */
int
icub_amplifier_reset(struct icub_node *node,
    enum icub_regset rs, uint8_t channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_AMPLIFIER_RESET;
	msg.data[1] = (rs << 4) | channel;
	return send_command(node, &msg, false);
}

/**
 * ask the range of the gain supported by the amplifier.
 *
 * page 63
 */
int
icub_amplifier_range_of_gain_get(struct icub_node *node,
    enum icub_regset rs, uint8_t channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_AMPLIFIER_RANGE_OF_GAIN_GET;
	msg.data[1] = (rs << 4) | channel;
	return send_command(node, &msg, true);
}

/**
 * ask the range of the offset supported by the amplifier.
 *
 * page 63
 */
int
icub_amplifier_range_of_offset_get(struct icub_node *node,
    enum icub_regset rs, uint8_t channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_AMPLIFIER_RANGE_OF_OFFSET_GET;
	msg.data[1] = (rs << 4) | channel;
	return send_command(node, &msg, true);
}

/**
 * retrieve the pair gain-offset from an amplifier.
 *
 * page 65
 */
int
icub_amplifier_gainoffset_get(struct icub_node *node,
    enum icub_regset rs, uint8_t channel)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 2;
	msg.data[0] = ICUB_AMPLIFIER_GAINOFFSET_GET;
	msg.data[1] = (rs << 4) | channel;
	return send_command(node, &msg, true);
}

/**
 * set the pair gain-offset from an amplifier.
 *
 * page 66
 */
int
icub_amplifier_gainoffset_set(struct icub_node *node,
    enum icub_regset rs, uint8_t channel,
    uint8_t mode, float gain, uint16_t offset)
{
	CAN_MSG msg;
	uint16_t g;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 7;
	msg.data[0] = ICUB_AMPLIFIER_GAINOFFSET_SET;
	msg.data[1] = (rs << 4) | channel;
	msg.data[2] = mode;
	g = (uint16_t)(gain * 100);
	msg.data[3] = g & 0x00ff;
	msg.data[4] = (g & 0xff00) >> 8;
	msg.data[5] = offset & 0x00ff;
	msg.data[6] = (offset & 0xff00) >> 8;
	return send_command(node, &msg, false);
}

/**
 * run auto-calibration for offset
 *
 * page 67
 */
int
icub_amplifier_offset_autocalib(struct icub_node *node,
    enum icub_regset rs, uint8_t channel,
    uint16_t target, uint16_t tolerance, uint8_t nSample)
{
	CAN_MSG msg;

	msg.id = ICUB_MSG_ID(ICUB_CLS_POLLING_ANALOG_SENSOR, 0, node->addr);
	msg.len = 8;
	msg.data[0] = ICUB_AMPLIFIER_GAINOFFSET_SET;
	msg.data[1] = (rs << 4) | channel;
	msg.data[2] = 0;	/* XX */
	msg.data[3] = target & 0x00ff;
	msg.data[4] = (target & 0xff00) >> 8;
	msg.data[5] = tolerance & 0x00ff;
	msg.data[6] = (tolerance & 0xff00) >> 8;
	msg.data[7] = nSample;
	return send_command(node, &msg, true);
}
	
      
