/*
 * Copyright (c) 2020-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <string.h>

#include <icub/icub.h>
#include "icub_protocol.h"
#include "internal.h"

int
icub_get_strain2_data(struct icub_node *node, struct strain2_data *data)
{
	struct strain2_data *strain;

	if (node->board_type != ICUB_BOARD_STRAIN2)
		return -1;

	strain = &(node->strain2);
	pthread_mutex_lock(&node->mutex);
	memcpy(data, strain, sizeof(struct strain2_data));
	pthread_mutex_unlock(&node->mutex);
	return 0;
}

void
icub_print_strain2_config(struct icub_node *node)
{
	struct strain2_data data, *strain = &data;
	int i, j;

	if (icub_get_strain2_data(node, strain) < 0)
		return;

	printf("serial: %s\n", strain->serial);
	printf("current regulation set: %d - validity %d\n",
		    strain->regset, strain->validity);
	for (i = 0; i < 6; i++) {
		printf("[");
		for (j = 0; j < 6; j++)
			printf("%8.4lf ", data.matrix[i][j]);
		printf("]\n");
	}
	printf("full scale: ");
	for (i = 0; i < 6; i++)
		printf("%.2lf ", data.fullscale[i]);
	printf("\ntare: ");
	for (i = 0; i < 6; i++)
		printf("%.2lf ", data.tare[i]);
	printf("\ngain: ");
	for (i = 0; i < 6; i++)
		printf("%.2lf ", data.gain[i]);
	printf("\noffset: ");
	for (i = 0; i < 6; i++)
		printf("%04x ", data.offset[i]);
	printf("\n");
}

void
icub_print_strain2_data(struct icub_node *node)
{
	struct strain2_data data, *strain = &data;

	if (icub_get_strain2_data(node, strain) < 0)
		return;

	switch (strain->mode) {
	case ICUB_STRAIN_MODE_TX_UNCALIBRATED:
		printf("force [uncalibrated] %lf %lf %lf\n",
		    strain->force[0],
		    strain->force[1], strain->force[2]);
		printf("torque [uncalibrated] %lf %lf %lf\n",
		    strain->torque[0],
		    strain->torque[1], strain->torque[2]);
		break;
	case ICUB_STRAIN_MODE_TX_CALIBRATED:
		printf("force [calibrated] %lf %lf %lf\n",
		    strain->force[0] * strain->fullscale[0], 
		    strain->force[1] * strain->fullscale[1],
		    strain->force[2] * strain->fullscale[2]);
		printf("torque [calibrated] %lf %lf %lf\n",
		    strain->torque[0] * strain->fullscale[3],
		    strain->torque[1] * strain->fullscale[4],
		    strain->torque[2] * strain->fullscale[5]);
		break;
	}
	printf("temperature %lf\n", strain->temperature);
	printf("IMU triple:");
	printf("acc %lf %lf %lf\n", strain->accel[0],
	    strain->accel[1], strain->accel[2]);
	printf("mag %lf %lf %lf\n", strain->mag[0],
	    strain->mag[1], strain->mag[2]);
	printf("gyr %lf %lf %lf\n", strain->gyro[0],
	    strain->gyro[1], strain->gyro[2]);
	printf("eul %lf %lf %lf\n", strain->euler[0],
	    strain->euler[1], strain->euler[2]);
	printf("lia %lf %lf %lf\n", strain->lia[0],
	    strain->lia[1], strain->lia[2]);
	printf("grv %lf %lf %lf\n", strain->grv[0],
	    strain->grv[1], strain->grv[2]);
	printf("quaternion %lf %lf %lf %lf\n", strain->q[0],
	    strain->q[1], strain->q[2], strain->q[3]);
	printf("imu_status seq %d gyr %d acc %d mag %d "
	    "duration %d ms\n", strain->seq,
	    strain->gyr_calib, strain->acc_calib, strain->mag_calib,
	    strain->acq_duration);
	printf("\n");
}
