/*
 * Copyright (c) 2020-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdint.h>
#include <stdio.h>

#include <icub/icub.h>

#include "socketcan.h"
#include "icub_protocol.h"
#include "internal.h"
#include "q15.h"

/**
 * Decodes messages received from Bootloader mode commands
 */
static int
decode_cmd_frame(struct icub_dev *dev, struct CAN_MSG *msg)
{
	struct icub_node *node = NULL;
	uint8_t *data = msg->data;
	int len = msg->len;
	int src = (msg->id & 0xf0) >> 4;
	int seq, rv = 0;

	if (src > 0 && src < 15)
		node = &(dev->nodes[src]);

	pthread_mutex_lock(&node->mutex);
	switch (data[0]) {
	case ICUB_CMD_BROADCAST: /* discover - page 19 */
		switch (len) {
		case 4:			/* bootloader mode */
			node->board_type = data[1];
			node->bl_maj = data[2];
			node->bl_min = data[3];
			fprintf(stderr, "Node %d is in bootloader mode!!\n",
			    src);
			break;
		case 5:			/* application mode */
			node->board_type = data[1];
			node->ap_maj = data[2];
			node->ap_min = data[3];
			node->ap_build = data[4];
			break;
		default:		/* invalid */
			rv=-1;
			break;
		}
		rv++;
		break;
	case ICUB_CMD_GET_ADDITIONAL_INFO: /* page 20 */
		if (len != 5) {
			rv = -1;
			break;
		}
		seq = data[1];
		if (seq < 1 || seq > 8) { /* invalid sequence number */
			rv = -1;
			break;
		}
		if (seq == 8) {
			/* full burst received -- assuming they are ordered */
			node->reply_pending = false;
			pthread_cond_signal(&node->cond);
			/* XXX store info & check if full string was received */
			rv++;
		}
		break;
	case ICUB_CMD_BOARD:	/* page 23 */
		if (len != 1) {
			rv = -1;
			break;
		}
		rv++;
		break;
	case ICUB_CMD_DATA:	/* payload - page 25*/
		if (len != 2) {
			rv = -1;
			break;
		}
		node->fw_ok = (data[1] == 1 ? true : false);
		rv++;
		break;
	case ICUB_CMD_START:	/* page 26 */
		if (len != 2)  {
			rv = -1;
			break;
		}
		node->fw_ok = (data[1] == 1 ? true : false);
		rv++;
		break;
	case ICUB_CMD_END:	/* page 27 */
		if (len != 2) {
			rv = -1;
			break;
		}
		node->fw_ok = (data[1] == 1 ? true : false);
		rv++;
		break;
	}
	pthread_mutex_unlock(&node->mutex);
	return rv;
}

/**
 * Decodes messages received from motor control polling commands
 *
 * here we only decode the GET_FIRMWARE_VERSION message
 * because it can be also sent by sensor boards like the strain2
 *
 * page 29
 */
static int
decode_polling_motor_control(struct icub_dev *dev, struct CAN_MSG *msg)
{
	uint8_t *data = msg->data;
	uint8_t src;
	int len = msg->len;
	struct icub_node *node;
	int rv = 0, board, compat;

	src = (msg->id & 0xf0) >> 4;
	node = &(dev->nodes[src]);
	pthread_mutex_lock(&node->mutex);
	switch (data[0]) {
	case ICUB_MOTOR_GET_FIRMWARE_VERSION:
		node->reply_pending = false;
		if (len != 8) {
			rv = -1;
			break;
		}
		board = data[1];
		if (board != node->board_type) {
			rv = -1;
			break;
		}
		node->ap_maj = data[2];
		node->ap_min = data[3];
		node->ap_build = data[4];
		node->can_maj = data[5];
		node->can_min = data[6];
		compat = data[7];
		if (!compat) {
			fprintf(stderr, "GET_FIRMWARE_VERSION: "
			    "incompatible CAN protocol version node %d %d.%d\n",
			    src, node->can_maj, node->can_min);
			rv = -1;
			break;
		}
		rv++;
		break;
	default:
		rv = -1;
		break;
	}
	pthread_cond_signal(&node->cond);
	pthread_mutex_unlock(&node->mutex);
	return rv;
}

/**
 * Decodes messages received from analog sensors polling commands
 */
static int
decode_polling_analog_sensor(struct icub_dev *dev, struct CAN_MSG *msg)
{
	uint8_t *data = msg->data;
	uint8_t src;
	int len = msg->len;
	struct icub_node *node;
	struct strain2_data *strain;
	int rv = 0, ch, fs, board, compat, row, col;
	enum icub_regset regset;

	src = (msg->id & 0xf0) >> 4;
	node = &(dev->nodes[src]);

	if (node->board_type != ICUB_BOARD_STRAIN &&
	    node->board_type != ICUB_BOARD_STRAIN2)
		return -1;

	strain = &(node->strain2);
	pthread_mutex_lock(&node->mutex);
	switch (data[0]) {
	case ICUB_GET_FIRMWARE_VERSION: /* page 33 */
		node->reply_pending = false;
		if (len != 8) {
			rv = -1;
			break;
		}
		board = data[1];
		if (board != node->board_type) {
			rv = -1;
			break;
		}
		node->ap_maj = data[2];
		node->ap_min = data[3];
		node->ap_build = data[4];
		node->can_maj = data[5];
		node->can_min = data[6];
		compat = data[7];
		if (!compat) {
			fprintf(stderr, "GET_FIRMWARE_VERSION: "
			    "incompatible CAN protocol version node %d %d.%d\n",
			    src, node->can_maj, node->can_min);
			rv = -1;
			break;
		}
		rv++;
		break;
	case ICUB_GET_FULLSCALE: /* page 36 */
		node->reply_pending = false;
		if (len != 4) {
			rv = -1;
			break;
		}
		regset = (data[1] & 0xf0) >> 4;
		/* only update values for current regulation set */
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			ch = (data[1] & 0x0f);
			fs = int16(data[2], data[3]); /* big endian (sic) */
			strain = &(node->strain2);
			strain->fullscale[ch] = fs;
		}
		rv++;
		break;
	case ICUB_GET_ADC:	/* page 37 */
		node->reply_pending = false;
		if (len != 5) {
			rv = -1;
			break;
		}
		ch = data[1];
		strain = &(node->strain2);
		strain->mode = data[2];
		if (ch < 3)
			strain->force[ch] = u16_to_float(data[3], data[4]);
		else
			strain->torque[ch-3] = u16_to_float(data[3], data[4]);
		rv++;
		break;
	case ICUB_IMU_CONFIG_GET: /* page 40 */
		node->reply_pending = false;
		strain->imu_config_mask = data[1] + (data[2] << 8);
		strain->imu_fusion_mode = data[3];
		rv++;
		break;
	case ICUB_THERMOMETER_CONFIG_GET: /* page 43 */
		node->reply_pending = false;
		rv++;
		break;
	case ICUB_REGULATIONSET_GET: /* page 50 */
		node->reply_pending = false;
		strain->regset = (data[1] & 0xf0) >> 4;
		strain->validity = data[1] & 0x0f;
		rv++;
		break;
	case ICUB_GET_MATRIX_RC: /* page 52 */
		node->reply_pending = false;
		rv++;
		regset = (data[1] & 0xf0) >> 4;
		/* only update values for current regulation set */
		if (regset != ICUB_REGSET_INUSE && regset != strain->regset)
			break;
		row = data[1] & 0x0f;
		col = data[2] & 0x0f;
		strain->matrix[row][col] =
                  q15_to_float(int16(data[3], data[4]));
		break;
	case ICUB_GET_SERIAL_NO: /* page 55 */
		node->reply_pending = false;
		rv++;
		for (ch = 0; ch < len - 1; ch++)
			strain->serial[ch] = data[ch+1];
		strain->serial[len-1] = '\0';
		break;
	case ICUB_GET_CH_DAC:	/* page 57 */
		node->reply_pending = false;
		rv++;
		regset = (data[1] & 0xf0) >> 4;
		ch = data[1] & 0x0f;
		/* only update values for current regulation set */
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			strain->offset[ch] = int16(data[2], data[3]);
		}
		break;
	case ICUB_GET_CALIB_TARE: /* page 59 */
		node->reply_pending = false;
		rv++;

		regset = (data[1] & 0xf0) >> 4;
		ch = data[1] & 0x0f;
		/* only update values for current regulation set */
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			strain->tare[ch] =
                          q15_to_float(int16(data[2], data[3]));
		}
		break;
	case ICUB_GET_EEPROM_STATUS:
		node->reply_pending = false;
		rv++;
		strain->eeprom_equal = data[1];
		break;
	case ICUB_SAVE2EE:
		node->reply_pending = false;
		rv++;
		strain->eeprom_ok = data[1];
		break;
	case ICUB_AMPLIFIER_RANGE_OF_GAIN_GET: /* page 63 */
		node->reply_pending = false;
		rv++;
		regset = (data[1] & 0xf0) >> 4;
		ch = data[1] & 0x0f;
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			strain->gain_min[ch] = int16(data[3], data[2]) / 100.0;
			strain->gain_max[ch] = int16(data[5], data[4]) / 100.0;
		}
		break;
	case ICUB_AMPLIFIER_RANGE_OF_OFFSET_GET: /* page 64 */
		node->reply_pending = false;
		rv++;
		regset = (data[1] & 0xf0) >> 4;
		ch = data[1] & 0x0f;
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			strain->offset_min[ch] = int16(data[3], data[2]);
			strain->offset_max[ch] = int16(data[5], data[4]);
		}
		break;
	case ICUB_AMPLIFIER_GAINOFFSET_GET: /* page 64 */
		node->reply_pending = false;
		rv++;
		regset = (data[1] & 0xf0) >> 4;
		ch = data[1] & 0x0f;
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			strain->gain[ch] = int16(data[3], data[2]) / 100.0;
			strain->offset[ch] = int16(data[5], data[4]);
		}
		break;
	case ICUB_AMPLIFIER_OFFSET_AUTOCALIB: /* page 67 */
		node->reply_pending = false;
		rv++;
		regset = (data[1] & 0xf0) >> 4;
		ch = data[1] & 0x0f;
		if (regset == ICUB_REGSET_INUSE || regset == strain->regset) {
			strain->chmsk = data[2];
			strain->algmsk = data[3];
			strain->memsk = data[4];
			/* data[5] is reserved for future use */
			strain->mae = int16(data[7], data[6]);
		}
		break;
	} /* switch */
	pthread_cond_signal(&node->cond);
	pthread_mutex_unlock(&node->mutex);
	return rv;
}

/**
 * Decodes periodic messages from analog sensors
 */
static int
decode_periodic_analog_sensor(struct icub_dev *dev, struct CAN_MSG *msg)
{
	uint8_t *data = msg->data;
	uint8_t src;
	struct icub_node *node;
	struct strain2_data *strain;
	int rv = 0;
	int typ;

	src = (msg->id & 0xf0) >> 4;
	typ = (msg->id & 0x0f);
	node = &(dev->nodes[src]);

	if (node->board_type != ICUB_BOARD_STRAIN &&
	    node->board_type != ICUB_BOARD_STRAIN2)
		return -1;

	strain = &(node->strain2);
	pthread_mutex_lock(&node->mutex);
	switch (typ) {
	case ICUB_UNCALIBFORCE_VECTOR_DEBUGMODE: /* page 81 */
	case ICUB_UNCALIBTORQUE_VECTOR_DEBUGMODE: /* page 82 */
		/* ignore them for now */
		rv++;
		break;
	case ICUB_FORCE_VECTOR:	/* page 83 */
		strain->force[0] = u16_to_float(data[1], data[0]);
		strain->force[1] = u16_to_float(data[3], data[2]);
		strain->force[2] = u16_to_float(data[5], data[4]);
		if (msg->len == 7)
			/* saturations */
			strain->force_saturation = data[6];
		else
			strain->force_saturation = 0;
		rv++;
		break;
	case ICUB_TORQUE_VECTOR: /* page 84 */
		strain->torque[0] = u16_to_float(data[1], data[0]);
		strain->torque[1] = u16_to_float(data[3], data[2]);
		strain->torque[2] = u16_to_float(data[5], data[4]);
		if (msg->len == 7)
			/* saturations */
			strain->torque_saturation = data[6];
		else
			strain->torque_saturation = 0;
		rv++;
		break;
	case ICUB_THERMOMETER_MEASURE: /* page 86 */
		strain->temperature = int16(data[2], data[1]) * 0.1;
		rv++;
		break;
	default:
		rv = -1;
	}
	pthread_mutex_unlock(&node->mutex);
	return rv;
}

/**
 * Decodes periodic messages from inertial sensors
 */
static int
decode_periodic_inertial_sensors(struct icub_dev *dev, struct CAN_MSG *msg)
{
	uint8_t *data = msg->data;
	uint8_t src;
	struct icub_node *node;
	struct strain2_data *strain;
	int typ, rv = 0;

	src = (msg->id & 0xf0) >> 4;
	typ = (msg->id & 0x0f);
	node = &(dev->nodes[src]);

	if (node->board_type != ICUB_BOARD_STRAIN2)
		return -1;
	strain = &(node->strain2);

	pthread_mutex_lock(&node->mutex);
	switch (typ) {
	case ICUB_DIGITAL_GYROSCOPE: /* page 94 */
		/* marked obsolete - not decoded */
		rv++;
		break;
	case ICUB_DIGITAL_ACCELEROMETER: /* page 95 */
		/* marked obsolet - not decoded */
		rv++;
		break;
	case ICUB_IMU_TRIPLE:	/* page 91 */
		strain->seq = data[0];
		switch (data[1]) {
		case ICUB_IMU_SENSOR_ACC:
			strain->accel[0] = int16(data[3], data[2]) * 0.01;
			strain->accel[1] = int16(data[5], data[4]) * 0.01;
			strain->accel[2] = int16(data[7], data[6]) * 0.01;
			break;
		case ICUB_IMU_SENSOR_MAG:
			strain->mag[0] = int16(data[3], data[2]) * 0.0625;
			strain->mag[1] = int16(data[5], data[4]) * 0.0625;
			strain->mag[2] = int16(data[7], data[6]) * 0.0625;
			break;
		case ICUB_IMU_SENSOR_GYR:
			strain->gyro[0] = int16(data[3], data[2]) * 0.0625;
			strain->gyro[1] = int16(data[5], data[4]) * 0.0625;
			strain->gyro[2] = int16(data[7], data[6]) * 0.0625;
			break;
		case ICUB_IMU_SENSOR_EUL:
			strain->euler[0] = int16(data[3], data[2]) * 0.0625;
			strain->euler[1] = int16(data[5], data[4]) * 0.0625;
			strain->euler[2] = int16(data[7], data[6]) * 0.0625;
			break;
		case ICUB_IMU_SENSOR_LIA:
			strain->lia[0] = int16(data[3], data[2]) * 0.01;
			strain->lia[1] = int16(data[5], data[4]) * 0.01;
			strain->lia[2] = int16(data[7], data[6]) * 0.01;
			break;
		case ICUB_IMU_SENSOR_GRV:
			strain->grv[0] = int16(data[3], data[2]) * 0.01;
			strain->grv[1] = int16(data[5], data[4]) * 0.01;
			strain->grv[2] = int16(data[7], data[6]) * 0.01;
			break;
		}
		rv++;
		break;
	case ICUB_IMU_QUATERNION: /* page 92 */
		strain->q[0] = int16(data[1], data[0]) / 16384.0;
		strain->q[1] = int16(data[3], data[2]) / 16384.0;
		strain->q[2] = int16(data[5], data[4]) / 16384.0;
		strain->q[3] = int16(data[7], data[6]) / 16384.0;
		rv++;
		break;
	case ICUB_IMU_STATUS:	/* page 93 */
		strain->seq = data[0];
		strain->gyr_calib = data[1];
		strain->acc_calib = data[2];
		strain->mag_calib = data[3];
		strain->acq_duration = data[4] + (data[5] << 8)
		    + (data[6] << 16) + (data[7] << 24);
		rv++;
		break;
	default:
		break;
	}
	pthread_mutex_unlock(&node->mutex);
	return rv;
}

/**
 * Dispatch decoding according to the message class
 */
static int
decode_frame(struct icub_dev *dev, struct CAN_MSG *msg)
{
	int cls = (msg->id & 0x700) >> 8;

	switch (cls) {
	case ICUB_CLS_POLLING_MOTOR_CONTROL:
		return decode_polling_motor_control(dev, msg);
	case ICUB_CLS_POLLING_ANALOG_SENSOR:
		return decode_polling_analog_sensor(dev, msg);
	case ICUB_CLS_BOOTLOADER:
		return decode_cmd_frame(dev, msg);
	case ICUB_CLS_PERIODIC_ANALOG_SENSOR:
		return decode_periodic_analog_sensor(dev, msg);
	case ICUB_CLS_PERIODIC_INERTIAL_SENSORS:
		return decode_periodic_inertial_sensors(dev, msg);
	default:
		return 0;
	}
	return -1;
}

/**
 * Wait for next CAN message and decode it
 *
 * return -1 in case of error 0 otherwise.
 * If no can frame was available it's not an error
 */
int
icub_receive_and_decode(struct icub_dev *dev, int timeout)
{
	CAN_MSG msg;
	int rv;

	rv = dev->recv(dev->fd, &msg, timeout);
	if (rv  <= 0)
		return rv;
	return decode_frame(dev, &msg);
}
