/*
 * Copyright (c) 2020-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>

#include <icub/icub.h>

#include "socketcan.h"
#include "icub_protocol.h"
#include "internal.h"

struct icub_dev *
icub_init(const char *name)
{
	struct icub_dev *dev;

	dev = (struct icub_dev *)malloc(sizeof(struct icub_dev));
	if (dev == NULL) {
		return NULL;
	}

	dev->fd = socketcanInit(name);
	if (dev->fd >= 0) {
          dev->send = socketcanTransmitMsg;
          dev->recv = socketcanReceiveMsgWait;
          dev->end = socketcanEnd;
        } else {
		dev->fd =
		  ttyInit(name, 0/*XXX allow passing a specific baud rate*/);
		if (dev->fd >= 0) {
			dev->send = ttyTransmitMsg;
			dev->recv = ttyReceiveMsgWait;
			dev->end = ttyEnd;
		} else {
			free(dev);
			return NULL;
		}
	}

	for (int i = 0; i < ICUB_MAX_NODES; i++) {
		struct icub_node *n = &dev->nodes[i];

		pthread_mutex_init(&n->mutex, NULL);
		pthread_cond_init(&n->cond, NULL);
		n->reply_pending = false;
		/* mark all nodes as unknown */
		n->board_type = 0xff;
		n->parent = dev;
		n->addr = i;
	}
	return dev;
}

int
icub_probe(struct icub_dev *dev)
{
	icub_discover(dev);
	while (icub_receive_and_decode(dev, 300) == 1)
		;
	return 0;
}

/**
 * return the parent device of a node
 */
struct icub_dev *
icub_dev(const struct icub_node *node)
{
	return node->parent;
}

/**
 * find node by type
 *
 * returns the node pointer or NULL if not found
 */
struct icub_node *
icub_find_node(struct icub_dev *dev, int board_type)
{
	int i;

	for (i = 1; i < ICUB_MAX_NODES; i++)
		if (dev->nodes[i].board_type == board_type)
			return &dev->nodes[i];
	return NULL;
}

/**
 * find node by address
 *
 * return the node pointer at given address
 */
struct icub_node *
icub_get_node(struct icub_dev *dev, int addr)
{
	return &dev->nodes[addr];
}

/**
 * return the CAN address of a node
 */
int
icub_can_address(const struct icub_node *node)
{
	return node->addr;
}

static void
dump_node(const struct icub_node *n)
{
	printf("node %d: type %02x\n", n->addr, n->board_type);
	printf("  version: %d.%d.%d\n", n->ap_maj, n->ap_min, n->ap_build);
	printf("  CAN protocol version: %d.%d\n", n->can_maj, n->can_min);
}

void
icub_dump(const struct icub_dev *dev)
{
	const struct icub_node *n;
	int i;

	for (i = 1; i < ICUB_MAX_NODES; i++) {
		n = &dev->nodes[i];
		if (n->board_type != ICUB_BOARD_UNKNOWN)
			dump_node(n);
	}
}

void
icub_end(struct icub_dev *dev)
{
	dev->end(dev->fd);
	free(dev);
}
