/*
 * Copyright (c) 2020 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _ICUB_PROTOCOL_H
#define _ICUB_PROTOCOL_H

/*
 * Contants used by the iCub Sensor boards on the CAN communication bus
 *
 * These constants are base on;
 *  IIT Technical Specification Document
 *  TSD-HSP-CAN-protocol-sensor-boards v0.3
 */

/*
 * CAN Interface version
 *
 * checked with the GET_FIRWARE_VERSION command
 */
#define ICUB_CAN_PROTO_MAJOR 2
#define ICUB_CAN_PROTO_MINOR 0

/* node addresses 4 bits - pages 4 & 6 */
/*    0 -> CAN controller */
/* 1-14 -> node           */
/*   15 -> broadcast      */
#define ICUB_MAX_NODES 14
#define ICUB_NODE_EVERYONE 15

/* CAN Message Classes - 3 bits - pages 13-14 */
#define ICUB_CLS_POLLING_MOTOR_CONTROL 0x0
#define ICUB_CLS_PERIODIC_MOTOR_CONTROL 0x1
#define ICUB_CLS_POLLING_ANALOG_SENSOR 0x2
#define ICUB_CLS_PERIODIC_ANALOG_SENSOR 0x3
#define ICUB_CLS_PERIODIC_SKIN_DATA 0x4
#define ICUB_CLS_PERIODIC_INERTIAL_SENSORS 0x5
/* reserved for future use 0x6 */
#define ICUB_CLS_BOOTLOADER 0x7

#define ICUB_IS_CMD_CLS(c) (				\
	    (c) == ICUB_CLS_POLLING_MOTOR_CONTROL ||	\
	    (c) == ICUB_CLS_POLLING_ANALOG_SENSOR ||	\
	    (c) == ICUB_CLS_BOOTLOADER)

/* assemble a CAN message ID from Class Source Destination - page 2*/
#define ICUB_MSG_ID(c, s, d) ((c) << 8 | (s) << 4 | (d))

/* Bootloader commands - page 17 */
#define ICUB_CMD_BOARD 0x00
#define ICUB_CMD_ADDRESS 0x01
#define ICUB_CMD_START 0x02
#define ICUB_CMD_DATA 0x03
#define ICUB_CMD_END 0x04
#define ICUB_CMD_GET_ADDITIONAL_INFO 0x0c
#define ICUB_CMD_SET_ADDITIONAL_INFO 0x0d
#define ICUB_CMD_SET_ADDRESS 0x32
#define ICUB_CMD_BROADCAST 0xff

/* Polling motor control commands - page 28 */
#define ICUB_SET_BOARD_ID 0x32
#define ICUB_MOTOR_GET_FIRMWARE_VERSION 0x5b

/* Polling analog sensors commands - page 30 */
#define ICUB_SET_MATRIX_RC 0x03
#define ICUB_SET_CH_DAC 0x04
#define ICUB_SET_TXMODE 0x07
#define ICUB_SET_CANDATARATE 0x08
#define ICUB_SAVE2EE 0x09
#define ICUB_GET_MATRIX_RC 0x0a
#define ICUB_GET_CH_DAC 0x0b
#define ICUB_GET_ADC 0x0c
#define ICUB_SET_RESOLUTION 0x10
#define ICUB_SET_CALIB_TARE 0x13
#define ICUB_GET_CALIB_TARE 0x14
#define ICUB_SET_FULLSCALE 0x17
#define ICUB_GET_FULLSCALE 0x18
#define ICUB_SET_SERIAL_NO 0x19
#define ICUB_GET_SERIAL_NO 0x1a
#define ICUB_GET_EEPROM_STATUS 0x1b
#define ICUB_GET_FIRMWARE_VERSION 0x1c
#define ICUB_AMPLIFIER_RESET 0x1d
#define ICUB_AMPLIFIER_RANGE_OF_GAIN_GET 0x1e
#define ICUB_AMPLIFIER_RANGE_OF_OFFSET_GET 0x1f
#define ICUB_AMPLIFIER_GAINOFFSET_GET 0x20
#define ICUB_AMPLIFIER_GAINOFFSET_SET 0x21
#define ICUB_AMPLIFIER_OFFSET_AUTOCALIB 0x22
#define ICUB_AMPLIFIER_PGA308_CFG1_GET 0x2a
#define ICUB_AMPLIFIER_PGA308_CFG1_SET 0x2b
#define ICUB_SET_BOARD_ADX 0x32
#define ICUB_IMU_CONFIG_GET 0x33
#define ICUB_IMU_CONFIG_SET 0x34
#define ICUB_IMU_TRANSMIT 0x35
#define ICUB_THERMOMETER_CONFIG_GET 0x38
#define ICUB_THERMOMETER_CONFIG_SET 0x39
#define ICUB_THERMOMETER_TRANSMIT 0x3a
#define ICUB_REGULATIONSET_SET 0x3d
#define ICUB_REGULATIONSET_GET 0x3e
#define ICUB_ACC_GYRO_SETUP 0x4f

/* TYP of streaming messages for periodic analog sensors - page 78 */
#define ICUB_USER_DEF 0x0
#define ICUB_POS_MEASURE 0x1
#define ICUB_UNCALIBFORCE_VECTOR_DEBUGMODE 0x8
#define ICUB_UNCALIBTORQUE_VECTOR_DEBUGMODE 0x9
#define ICUB_FORCE_VECTOR 0xa
#define ICUB_TORQUE_VECTOR 0xb
#define ICUB_HES0TO6 0xc
#define ICUB_HES7TO14 0xd
#define ICUB_THERMOMETER_MEASURE 0xe

/* TYP of streaming messages for periodic inertial data - page 90 */
#define ICUB_DIGITAL_GYROSCOPE 0x0
#define ICUB_DIGITAL_ACCELEROMETER 0x01
#define ICUB_IMU_TRIPLE 0x03
#define ICUB_IMU_QUATERNION 0x04
#define ICUB_IMU_STATUS 0x05

/* IMU stream data - page 91 */
#define	ICUB_IMU_SENSOR_ACC 0
#define	ICUB_IMU_SENSOR_MAG 1
#define	ICUB_IMU_SENSOR_GYR 2
#define	ICUB_IMU_SENSOR_EUL 3
#define	ICUB_IMU_SENSOR_QUA 4
#define	ICUB_IMU_SENSOR_LIA 5
#define	ICUB_IMU_SENSOR_GRV 6
#define	ICUB_IMU_SENSOR_STATUS 15
#define ICUB_IMU_SENSOR_NONE 16

#endif	/* _ICUB_PROTOCOL_H */
