/*
 * Copyright (c) 2020-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "icub-config.h"

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#ifdef HAVE_LOW_LATENCY_IOCTL
# include <sys/ioctl.h>
# include <linux/serial.h> /* for TIOCGSERIAL and ASYNC_LOW_LATENCY */
#endif

#ifdef __linux__
# include <libudev.h>
#endif

# include <socketcan.h>

static const char *usbSerialToTty(const char *serial);

static inline uint8_t nibtohex(uint8_t x);
static inline uint8_t hextonib(uint8_t x);


/* --- ttyInit ------------------------------------------------------------- */

/** Open a serial port, configure to the given baud rate (unless speed == 0) */

int _PRIVATE
ttyInit(const char *deviceName, uint32_t speed)
{
	const char *path;
	struct termios t;
	speed_t baud;
	int fd, e;

	/* select baud rate */
#ifndef B57600
# define B57600 57600U
#endif
#ifndef B115200
# define B115200 115200U
#endif
#ifndef B500000
# define B500000 500000U
#endif
#ifndef B1000000
# define B1000000 1000000U
#endif
#ifndef B2000000
# define B2000000 2000000U
#endif
	switch(speed) {
	case 0:		baud = 0; break;
	case 57600:	baud = B57600; break;
	case 115200:	baud = B115200; break;
	case 500000:	baud = B500000; break;
	case 1000000:	baud = B1000000; break;
	case 2000000:	baud = B2000000; break;

	default: errno = EINVAL; return -1;
	}

	/* try to match by serial first */
	path = usbSerialToTty(deviceName);
	if (path) deviceName = path;

	/* open blocking */
	fd = open(deviceName, O_RDWR | O_NOCTTY);
	if (fd < 0) return fd;
	if (!isatty(fd)) {
	  errno = ENOTTY;
	  goto err;
	}

	/* configure line discipline: cannonical mode, \r EOL */
	if (tcgetattr(fd, &t)) goto err;

	t.c_iflag = IGNBRK;
	t.c_oflag = 0;
	t.c_lflag = ICANON;
	t.c_cflag &= ~(CSIZE | CSTOPB | PARENB | PARODD);
	t.c_cflag |= CS8 | CREAD | CLOCAL | HUPCL;
	t.c_cc[VEOL] = '\r';

	if (speed) {
	  if (cfsetospeed(&t, baud)) goto err;
	  if (cfsetispeed(&t, baud)) goto err;
	}

	if (tcsetattr(fd, TCSANOW, &t)) goto err;

#ifdef HAVE_LOW_LATENCY_IOCTL
	/* Linux: enable low latency mode
	 * see /sys/bus/usb-serial/devices/.../latency_timer
	 */
	{
		struct serial_struct s;

		if (!ioctl(fd, TIOCGSERIAL, &s)) {
			if (!(s.flags & ASYNC_LOW_LATENCY)) {
				s.flags |= ASYNC_LOW_LATENCY;
				ioctl(fd, TIOCSSERIAL, &s);
			}
                }
	}
#endif

	/* switch to non-blocking */
	if (fcntl(fd, F_SETFL, O_NONBLOCK)) {
		goto err;
	}

	/* discard any pending data */
	tcflush(fd, TCIOFLUSH);

	/* slcan init: close, configure 1Mbit/s bus, open */
	static const char ostr[] = "C\rS8\rO\r";
	if (write(fd, ostr, sizeof(ostr)-1) <= 0)
		goto err;

	return fd;
err:
	e = errno;
	close(fd);
	errno = e;
	return -1;
}


/* --- ttyEnd --------------------------------------------------------------*/
/**
 * End the communication with the CAN bus controller
 */
void _PRIVATE
ttyEnd(int fd)
{
	static const char cstr[] = "C\r";
	if (write(fd, cstr, sizeof(cstr)-1) <= 0) /* ignore err */;
	close(fd);
}


/* --- ttyTransmitMsg ------------------------------------------------------ */

/** Send a message on the BUS */
int _PRIVATE
ttyTransmitMsg(int s, CAN_MSG *msg)
{
	uint8_t frame[22], *p, *w;
	ssize_t l;
	uint8_t i;

#ifdef CAN_BUS_DEBUG
	printf("-> "); canMsgShow(stdout, msg);
#endif

	/* 'transmit' frame */
	p = frame;
	*p++ = 't';

	/* id */
	*p++ = nibtohex(msg->id >> 8);
	*p++ = nibtohex(msg->id >> 4);
	*p++ = nibtohex(msg->id);

	/* len */
	*p++ = nibtohex(msg->len);

	/* data */
	for(i = 0; i < msg->len; i++) {
		*p++ = nibtohex(msg->data[i] >> 4);
		*p++ = nibtohex(msg->data[i]);
	}

	/* send */
	*p++ = '\r';
	w = frame;
	for(w = frame; w < p; w += l) {
		l = write(s, w, p - w);
		if (l < 0) return -1;
	}

	return 0;
}


/* --- ttyReceiveMsgWait -------------------------------------------------- */
/**
 * Wait for a message and receive it
 *
 * @param timeo number of milli-seconds to wait
 */
int _PRIVATE
ttyReceiveMsgWait(int s, CAN_MSG *msg, int timeo)
{
	struct timeval start, end;
	uint8_t frame[24], *p;
	struct pollfd pfd;
	ssize_t nbytes;
        uint8_t i;
	int rc;

	gettimeofday(&start, NULL);
	for(;;) {
		/* update timeout */
		gettimeofday(&end, NULL);
		timeo -= (end.tv_sec - start.tv_sec) * 1000;
		timeo -= (end.tv_usec - start.tv_usec) / 1000;
		if (timeo <= 0) return 0;

		pfd.fd = s;
		pfd.events = POLLIN;
		for (;;) {
			rc = poll(&pfd, 1, timeo);
			if (rc == -1 && errno == EINTR)
				continue;
			if (rc <= 0)
				return rc;
			break;
		}

		nbytes = read(s, frame, sizeof(frame));
		if (nbytes == 0)
			return 0;
		if (nbytes < 0) {
			if (errno == EAGAIN)
				return 0;
#ifdef CAN_BUS_DEBUG
			warn("ttyReceiveMsgWait: read");
#endif
			return -1;
		}

		switch(frame[0]) {
		case '\r':
#ifdef CAN_BUS_DEBUG
			printf("<- ack\n");
#endif
			break;

		case 't':
			p = &frame[1];

			/* id */
			msg->id = hextonib(*p++) << 8;
			msg->id |= hextonib(*p++) << 4;
			msg->id |= hextonib(*p++);

			/* len */
			msg->len = hextonib(*p++);
			if (msg->len > 8) /* bad message */ break;

			/* data */
			for(i = 0; i < msg->len; i++) {
				msg->data[i] = hextonib(*p++) << 4;
				msg->data[i] |= hextonib(*p++);
			}

#ifdef CAN_BUS_DEBUG
			printf("<- "); canMsgShow(stdout, msg);
#endif
			return 1;
		}
	}

	/*NOTREACHED*/
	return 0;
}


/* --- usbSerialToTty ------------------------------------------------------ */

/** Return a tty device matching the "serial" string */

static const char *
usbSerialToTty(const char *serial)
{
#ifdef __linux__
	struct udev *udev;
	struct udev_enumerate *scan = NULL;
	struct udev_list_entry *ttys, *tty;
	struct udev_device *dev = NULL, *p;
	const char *path = NULL;
	const char *s;

	udev = udev_new();
	if (!udev) return NULL;

	/* iterate over tty devices */
	scan = udev_enumerate_new(udev);
	if (udev_enumerate_add_match_subsystem(scan, "tty")) goto done;
	if (udev_enumerate_scan_devices(scan)) goto done;

	ttys = udev_enumerate_get_list_entry(scan);
	udev_list_entry_foreach(tty, ttys) {
		const char *sysfs, *pserialstr, *pifacestr, *pnifacestr;
		int suf, piface, pniface;

		/* get sysfs entry for the device and create a corresponding
		 * udev_device */
		if (dev) udev_device_unref(dev);
		sysfs = udev_list_entry_get_name(tty);
		dev = udev_device_new_from_syspath(udev, sysfs);

		/* iterate over parents and look for serial or bInterfaceNumber
		 * for usb */
		pserialstr = pifacestr = pnifacestr = NULL;
		for(p = udev_device_get_parent(dev);
                    p; p = udev_device_get_parent(p)) {
			if (!pserialstr)
				pserialstr = udev_device_get_sysattr_value(
                                  p, "serial");
			if (!pifacestr)
				pifacestr = udev_device_get_sysattr_value(
                                  p, "bInterfaceNumber");
			if (!pnifacestr)
				pnifacestr = udev_device_get_sysattr_value(
                                  p, "bNumInterfaces");
		}
		piface = pifacestr ? atoi(pifacestr) : 0;
		pniface = pnifacestr ? atoi(pnifacestr) : 0;

		/* check match */
		if (!pserialstr) continue; /* no device serial */
		if (strstr(serial, pserialstr) != serial)
                  continue; /* no prefix match */
		s = serial + strlen(pserialstr); /* advance after prefix */
		if (pniface > 1 && *s == '\0') { /* ambiguous match */
			warnx("serial %s matches %s.%d but is ambiguous",
                                serial, pserialstr, piface);
			continue;
		}
		if (sscanf(s, ".%d", &suf) != 1) suf = -1;
		if (pniface < 2 && *s && suf != piface)
                  continue; /* no suffix match */
		if (pniface > 1 && suf != piface)
                  continue; /* no suffix match */

		/* got a match, return the tty path */
		path = strdup(udev_device_get_devnode(dev)); /* XXX leaks */
		break;
	}
	if (dev) udev_device_unref(dev);

done:
	if (scan) udev_enumerate_unref(scan);
	if (udev) udev_unref(udev);
	return path;

#else
	(void)serial; /* -Wunused-parameter */
	return NULL; /* not implemented */
#endif
}


/* --- ascii / bin conversions --------------------------------------------- */

static inline uint8_t
nibtohex(uint8_t x)
{
	return (x & 0xf) + (((x & 0xf) < 0xa) ? '0' : 'A' - 0xa);
}

static inline uint8_t
hextonib(uint8_t x)
{
	return x - ((x < 'A') ? '0' : (x < 'a') ? 'A' - 0xa : 'a' - 0xa);
}
