/*
 * Copyright (c) 2014-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "icub-config.h"

#include <err.h>
#include <errno.h>
#include <poll.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#ifdef HAVE_LINUX_CAN_H
# include <linux/can.h>
#endif

#include "socketcan.h"

// #define CAN_BUS_DEBUG 1

/**
 ** Interface with the linux socket CAN layer
 **/

/*----------------------------------------------------------------------*/
/**
 * Initialize the communication with the CAN bus controller
 */
int _PRIVATE
socketcanInit(const char *deviceName)
{
#ifdef HAVE_LINUX_CAN_H
	struct ifreq ifr;
	struct sockaddr_can addr;
	int s;

	s = socket(PF_CAN, SOCK_RAW, CAN_RAW);
	snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s", deviceName);
	if (ioctl(s, SIOCGIFINDEX, &ifr) != 0) {
		warn("SIOCGIFINDEX %s", deviceName);
		return -1;
	}
	addr.can_family = AF_CAN;
	addr.can_ifindex = ifr.ifr_ifindex;

	if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		warn("bind");
		return -1;
	}
	return s;
#else
	errno = EINVAL;
	return -1;
#endif
}

/*----------------------------------------------------------------------*/
/**
 * End the communication with the CAN bus controller
 */
void _PRIVATE
socketcanEnd(int socket)
{
#ifdef HAVE_LINUX_CAN_H
	close(socket);
#endif
}

/*----------------------------------------------------------------------*/
/**
 * Send a message on the BUS
 */
int _PRIVATE
socketcanTransmitMsg(int s, CAN_MSG *msg)
{
#ifdef HAVE_LINUX_CAN_H
	ssize_t nbytes;
	struct can_frame frame;

#ifdef CAN_BUS_DEBUG
	printf("-> "); canMsgShow(stdout, msg);
#endif
	memset(&frame, 0, sizeof(struct can_frame));
	frame.can_id = msg->id;
	frame.can_dlc = msg->len;
	memcpy(&frame.data, msg->data, msg->len);

	nbytes = write(s, &frame, sizeof(struct can_frame));
#ifdef CAN_BUS_DEBUG
	if (nbytes < 0)
		warn("socketcanTransmitMsg: write");
#endif
	return nbytes == sizeof(struct can_frame) ? 0 : -1;
#else
	errno = EINVAL;
	return -1;
#endif
}

/*----------------------------------------------------------------------*/
/**
 * Wait for a message and receive it
 *
 * @param timeo number of milli-seconds to wait : -1 => wait forever
 */
int _PRIVATE
socketcanReceiveMsgWait(int s, CAN_MSG *msg, int timeo)
{
#ifdef HAVE_LINUX_CAN_H
	struct can_frame frame;
	struct pollfd pfd;
	ssize_t nbytes;
	int rc;

	if (timeo >= 0) {
		pfd.fd = s;
		pfd.events = POLLIN;
		for (;;) {
			rc = poll(&pfd, 1, timeo);
			if (rc == -1 && errno == EINTR)
				continue;
			if (rc <= 0)
				return rc;
			break;
		}
	}
	nbytes = read(s, &frame, sizeof(struct can_frame));
	if (nbytes < 0) {
#ifdef CAN_BUS_DEBUG
		warn("socketcanReceiveMsg: read");
#endif
		return -1;
	}

	/* paranoid check ... */
	if (nbytes < sizeof(struct can_frame)) {
#ifdef CAN_BUS_DEBUG
		warnx("socketcanReceiveMsg: read: incomplete CAN frame");
#endif
		return -1;
	}
	msg->id = frame.can_id;
	msg->len = frame.can_dlc;
	memcpy(msg->data, frame.data, frame.can_dlc);
#ifdef CAN_BUS_DEBUG
	printf("<- "); canMsgShow(stdout, msg);
#endif
	return 1;
#else
	errno = EINVAL;
	return -1;
#endif
}

/*----------------------------------------------------------------------*/
/**
 * Non-Blocking check for next message and receive it
 */
int _PRIVATE
socketcanReceiveMsg(int s, CAN_MSG *m)
{
	return socketcanReceiveMsgWait(s, m, 0);
}

/*----------------------------------------------------------------------*/
/**
 * Initialize a CAN message
 */
void _PRIVATE
canMsgSet(CAN_MSG *m, int id,
    int p1, int p2, int p3, int p4, int p5, int p6, int p7, int p8)
{
	m->id = id;
	m->len = 8;
	m->data[0] = p1;
	m->data[1] = p2;
	m->data[2] = p3;
	m->data[3] = p4;
	m->data[4] = p5;
	m->data[5] = p6;
	m->data[6] = p7;
	m->data[7] = p8;
}

/*----------------------------------------------------------------------*/
/**
 * Display the raw contents of a CAN message on stdout
 */
void _PRIVATE
canMsgShow(FILE *out, CAN_MSG *m)
{
	int i;

	fprintf(out, "ID: 0x%x ", m->id);
	fprintf(out, "len: %d ", m->len);
	fprintf(out, "data:");
	for (i = 0; i < m->len; i++) {
		fprintf(out, " 0x%02x", m->data[i]);
	}
	fprintf(out, "\n");
}
