/*
 * Copyright (c) 2020-2021 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <err.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <icub/icub.h>

int done = 0;

void *
icubReadTask(void *data)
{
	struct icub_node *node = data;
	struct icub_dev *dev = icub_dev(node);

	while (done == 0) {
		if (icub_receive_and_decode(dev, 200) < 0) {
			warn("error reading can bus");
			break;
		}
		icub_print_strain2_data(node);
	}
	return NULL;
}

void
icubStop(int sig)
{
	/* ask read task to terminate */
	done = 1;
}

int
main(int argc, char *argv[])
{
	char *name;
	struct icub_dev *dev;
	struct icub_node * node;
	int  i, j;
	pthread_t readTask;

	if (argc == 2)
		name = argv[1];
	else
		name = "can0";

	dev = icub_init(name);
	if (dev == NULL)
		err(2, "icub_init(%s)", name);

	/* probe device until one is found, with a timeout */
	for(i = 0; i < 10; i++) {
		if (icub_probe(dev) != 0)
			errx(2, "icub_probe");
		if (icub_find_node(dev, ICUB_BOARD_STRAIN2) != NULL)
			break;
	}
	if ((node = icub_find_node(dev, ICUB_BOARD_STRAIN2)) == NULL)
		errx(2, "can't find strain2 node");
	printf("found strain2 at %d\n", icub_can_address(node));
	if ((icub_sensor_get_firmware_version(node) != 0) ||
	    (icub_receive_and_decode(dev, 100) < 0))
		errx(2, "get_firmware_version");
	icub_dump(dev);

	signal(SIGINT, icubStop);

	pthread_create(&readTask, NULL, icubReadTask, node);

	icub_get_serial_number(node);
	icub_regulationset_get(node, ICUB_REGSET_VALIDITY_TEMPORARY);
	icub_set_calib_tare(node, ICUB_CALIB_MODE_EVERY_CHANNEL_NEGATIVE_OF_DAC,
                ICUB_REGSET_INUSE, 0, 0);
	for (i = 0; i < 6; i++) {
		icub_sensor_get_fullscale(node, ICUB_REGSET_INUSE, i);
		icub_get_calib_tare(node, ICUB_REGSET_INUSE, i);
	}
	for (i = 0; i < 6; i++)
		for (j = 0; j < 6; j++)
			icub_get_matrix_rc(node, ICUB_REGSET_INUSE, i, j);
	for (i = 0; i < 6; i++)
		icub_amplifier_gainoffset_get(node, ICUB_REGSET_INUSE, i);
	icub_sync(node);
	icub_print_strain2_config(node);
	/* set up data receiving */
	icub_sensor_set_can_data_rate(node, 100);
	icub_sensor_set_txmode(node, ICUB_STRAIN_MODE_TX_CALIBRATED);
	icub_sensor_thermometer_config_set(node, ICUB_THERMOMETER_ONE);
	icub_sensor_thermometer_transmit(node, 1);
	icub_sensor_imu_config_set(node, ICUB_IMU_SENSOR_MASK_ACC
	    | ICUB_IMU_SENSOR_MASK_MAG
	    | ICUB_IMU_SENSOR_MASK_GYR
	    | ICUB_IMU_SENSOR_MASK_EUL
	    | ICUB_IMU_SENSOR_MASK_QUA
	    | ICUB_IMU_SENSOR_MASK_LIA
	    | ICUB_IMU_SENSOR_MASK_GRV
	    | ICUB_IMU_SENSOR_MASK_STATUS);
	icub_sensor_imu_transmit(node, 200);
	/* let it run for 10s */
	sleep(10);
	/* stop data transmission */
	icubStop(0);
	icub_sensor_set_txmode(node, ICUB_STRAIN_MODE_TX_NONE);
	icub_sensor_thermometer_transmit(node, 0);
	icub_sensor_imu_transmit(node, 0);
	pthread_join(readTask, NULL);
	icub_end(dev);
	return 0;
}
